#  tests for sympy-1.10.1-py311h06a4308_0 (this is a generated file);
print('===== testing package: sympy-1.10.1-py311h06a4308_0 =====');
print('running run_test.py');
#  --- run_test.py (begin) ---
import sympy
from sympy import Rational


a = Rational(14, 4)
assert str(a) == '7/2'
#  --- run_test.py (end) ---

print('===== sympy-1.10.1-py311h06a4308_0 OK =====');
print("import: 'sympy'")
import sympy

print("import: 'sympy.algebras'")
import sympy.algebras

print("import: 'sympy.assumptions'")
import sympy.assumptions

print("import: 'sympy.assumptions.handlers'")
import sympy.assumptions.handlers

print("import: 'sympy.assumptions.predicates'")
import sympy.assumptions.predicates

print("import: 'sympy.assumptions.relation'")
import sympy.assumptions.relation

print("import: 'sympy.benchmarks'")
import sympy.benchmarks

print("import: 'sympy.calculus'")
import sympy.calculus

print("import: 'sympy.categories'")
import sympy.categories

print("import: 'sympy.codegen'")
import sympy.codegen

print("import: 'sympy.combinatorics'")
import sympy.combinatorics

print("import: 'sympy.concrete'")
import sympy.concrete

print("import: 'sympy.core'")
import sympy.core

print("import: 'sympy.core.benchmarks'")
import sympy.core.benchmarks

print("import: 'sympy.crypto'")
import sympy.crypto

print("import: 'sympy.diffgeom'")
import sympy.diffgeom

print("import: 'sympy.discrete'")
import sympy.discrete

print("import: 'sympy.external'")
import sympy.external

print("import: 'sympy.functions'")
import sympy.functions

print("import: 'sympy.functions.combinatorial'")
import sympy.functions.combinatorial

print("import: 'sympy.functions.elementary'")
import sympy.functions.elementary

print("import: 'sympy.functions.elementary.benchmarks'")
import sympy.functions.elementary.benchmarks

print("import: 'sympy.functions.special'")
import sympy.functions.special

print("import: 'sympy.functions.special.benchmarks'")
import sympy.functions.special.benchmarks

print("import: 'sympy.geometry'")
import sympy.geometry

print("import: 'sympy.holonomic'")
import sympy.holonomic

print("import: 'sympy.integrals'")
import sympy.integrals

print("import: 'sympy.integrals.benchmarks'")
import sympy.integrals.benchmarks

print("import: 'sympy.integrals.rubi'")
import sympy.integrals.rubi

print("import: 'sympy.integrals.rubi.parsetools'")
import sympy.integrals.rubi.parsetools

print("import: 'sympy.integrals.rubi.rubi_tests'")
import sympy.integrals.rubi.rubi_tests

print("import: 'sympy.integrals.rubi.rules'")
import sympy.integrals.rubi.rules

print("import: 'sympy.interactive'")
import sympy.interactive

print("import: 'sympy.liealgebras'")
import sympy.liealgebras

print("import: 'sympy.logic'")
import sympy.logic

print("import: 'sympy.logic.algorithms'")
import sympy.logic.algorithms

print("import: 'sympy.logic.utilities'")
import sympy.logic.utilities

print("import: 'sympy.matrices'")
import sympy.matrices

print("import: 'sympy.matrices.benchmarks'")
import sympy.matrices.benchmarks

print("import: 'sympy.matrices.expressions'")
import sympy.matrices.expressions

print("import: 'sympy.multipledispatch'")
import sympy.multipledispatch

print("import: 'sympy.ntheory'")
import sympy.ntheory

print("import: 'sympy.parsing'")
import sympy.parsing

print("import: 'sympy.parsing.autolev'")
import sympy.parsing.autolev

print("import: 'sympy.parsing.autolev._antlr'")
import sympy.parsing.autolev._antlr

print("import: 'sympy.parsing.c'")
import sympy.parsing.c

print("import: 'sympy.parsing.fortran'")
import sympy.parsing.fortran

print("import: 'sympy.parsing.latex'")
import sympy.parsing.latex

print("import: 'sympy.parsing.latex._antlr'")
import sympy.parsing.latex._antlr

print("import: 'sympy.physics'")
import sympy.physics

print("import: 'sympy.physics.continuum_mechanics'")
import sympy.physics.continuum_mechanics

print("import: 'sympy.physics.control'")
import sympy.physics.control

print("import: 'sympy.physics.hep'")
import sympy.physics.hep

print("import: 'sympy.physics.mechanics'")
import sympy.physics.mechanics

print("import: 'sympy.physics.optics'")
import sympy.physics.optics

print("import: 'sympy.physics.quantum'")
import sympy.physics.quantum

print("import: 'sympy.physics.units'")
import sympy.physics.units

print("import: 'sympy.physics.units.definitions'")
import sympy.physics.units.definitions

print("import: 'sympy.physics.units.systems'")
import sympy.physics.units.systems

print("import: 'sympy.physics.vector'")
import sympy.physics.vector

print("import: 'sympy.plotting'")
import sympy.plotting

print("import: 'sympy.plotting.intervalmath'")
import sympy.plotting.intervalmath

print("import: 'sympy.plotting.pygletplot'")
import sympy.plotting.pygletplot

print("import: 'sympy.polys'")
import sympy.polys

print("import: 'sympy.polys.agca'")
import sympy.polys.agca

print("import: 'sympy.polys.benchmarks'")
import sympy.polys.benchmarks

print("import: 'sympy.polys.domains'")
import sympy.polys.domains

print("import: 'sympy.polys.matrices'")
import sympy.polys.matrices

print("import: 'sympy.printing'")
import sympy.printing

print("import: 'sympy.printing.pretty'")
import sympy.printing.pretty

print("import: 'sympy.sandbox'")
import sympy.sandbox

print("import: 'sympy.series'")
import sympy.series

print("import: 'sympy.series.benchmarks'")
import sympy.series.benchmarks

print("import: 'sympy.sets'")
import sympy.sets

print("import: 'sympy.sets.handlers'")
import sympy.sets.handlers

print("import: 'sympy.simplify'")
import sympy.simplify

print("import: 'sympy.solvers'")
import sympy.solvers

print("import: 'sympy.solvers.benchmarks'")
import sympy.solvers.benchmarks

print("import: 'sympy.solvers.diophantine'")
import sympy.solvers.diophantine

print("import: 'sympy.solvers.ode'")
import sympy.solvers.ode

print("import: 'sympy.stats'")
import sympy.stats

print("import: 'sympy.stats.sampling'")
import sympy.stats.sampling

print("import: 'sympy.strategies'")
import sympy.strategies

print("import: 'sympy.strategies.branch'")
import sympy.strategies.branch

print("import: 'sympy.tensor'")
import sympy.tensor

print("import: 'sympy.tensor.array'")
import sympy.tensor.array

print("import: 'sympy.tensor.array.expressions'")
import sympy.tensor.array.expressions

print("import: 'sympy.testing'")
import sympy.testing

print("import: 'sympy.unify'")
import sympy.unify

print("import: 'sympy.utilities'")
import sympy.utilities

print("import: 'sympy.utilities._compilation'")
import sympy.utilities._compilation

print("import: 'sympy.utilities.mathml'")
import sympy.utilities.mathml

print("import: 'sympy.vector'")
import sympy.vector

